////////////////////////////////////////////////////////////////////// 
// relativeFinder/Main.cpp 
// (c) 2000-2010 Goncalo Abecasis
// 
// This file is distributed as part of the Goncalo source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile Goncalo.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday May 19, 2010
// 
 
#include "RelativeParameters.h"
#include "MerlinCluster.h"
#include "MerlinCore.h"
#include "MerlinError.h"
#include "Houdini.h"
#include "Manners.h"
#include "AutoFit.h"
#include "Error.h"

#include "LongInt.h"

void ShowBanner();

int main(int argc, char * argv[])
   {
   ShowBanner();

   SetupCrashHandlers();

   FuzzyInheritanceTree::perAlleleError = 0.002;

   RelativeParameters pl;

   pl.Read(argc, argv);
   pl.Status();

#ifdef __CHROMOSOME_X__
   PedigreeGlobals::chromosomeX = true;
#endif

   Pedigree ped;

   ped.Prepare(pl.dataFile);
   ped.Load(pl.pedigreeFile);
   ped.LoadMarkerMap(pl.mapFile, true /* filter out markers not in pedigree */);
   ped.LoadAlleleFrequencies(pl.freqFile);

   Pedigree testCases;

   testCases.Prepare("testCases.dat");
   testCases.Load("testCases.ped");

   // Estimate allele frequencies
   if (pl.alleleFrequencies == FREQ_ML)
      MarkerCluster::EstimateAlleleFrequencies(ped, MerlinCore::filePrefix + "-freqs.log");
   else
      {
      ped.EstimateFrequencies(pl.alleleFrequencies);

      // Recode alleles so more frequent alleles have lower allele numbers internally
      ped.LumpAlleles(0.0);
      }

   if (pl.trimPedigree)
      ped.Trim();

   clusters.LoadFromFile(pl.clusterFile, MerlinCore::filePrefix + "-clusters.log");
   clusters.ClusterByDistance(pl.clusterDistance * 0.01);
   clusters.ClusterByRsquared(ped, pl.clusterRsquared);
   clusters.EstimateFrequencies(ped, MerlinCore::filePrefix + "-cluster-freqs.log");
   clusters.FinishOutput();

   if (ped.markerCount < 1)
      error("The data set includes no genetic markers\n\n");

   if (ped.count < 1)
      error("Your pedigree must include at least two individuals!");

   if (ped.sexSpecificMap)
      error("Sex-specific maps may not be supported correctly");
//      Mantra::ignoreCoupleSymmetries = true;

   IntArray sortedMarkers;
   ped.SortMarkersInMapOrder(sortedMarkers, -1);

   // Count the number of pairs with genotype data in the pedigree
   int pairs = 0;
   for (int j = 0; j < ped.count; j++)
      {
      if (ped[j].ngeno == 0) continue;

      int last = !pl.withinFamilyPairs ? ped.count :
                 ped.FindFamily(ped[j].famid)->last + 1;

      for (int k = j + 1; k < last; k++)
         {
         if (ped[k].ngeno == 0) continue;

         pairs++;
         }
      }

   if (pairs < 1)
      error("There are no genotyped pairs in the pedigree");

   printf("Evaluating %d possible relationships for each of %d pairs\n",
          testCases.familyCount, pairs);

   if (pl.jobCount > pairs)
      pl.jobCount = pairs;

   if (pl.job < 1 || pl.jobCount < 1)
      error("Invalid job parameters (requested job %d of %d)", pl.job, pl.jobCount);

   if (pl.job > pl.jobCount)
      {
      printf("  ** Nothing to do in this batch job **\n\n");
      exit(0);
      }

   int jobStart = (int) ((long long) (pairs * (pl.job - 1)) / pl.jobCount);
   int jobEnd = (int) ((long long) (pairs * pl.job) / pl.jobCount);

   if (pl.jobCount > 1)
      printf("  Job %d of %d will evaluate %d pairs\n",
             pl.job, pl.jobCount, jobEnd - jobStart);

   float ** likelihoods = new float * [testCases.familyCount + 1];
   for (int i = 0; i < testCases.familyCount + 1; i++)
      {
      likelihoods[i] = (new float [jobEnd - jobStart]) - jobStart;

      for (int j = jobStart; j < jobEnd; j++)
         likelihoods[i][j] = 0.0;
      }

   int  next_chromosome  = -2;
   bool many_chromosomes = false;

   String prefix = MerlinCore::filePrefix;

   MerlinCore engine(testCases);
   engine.SetupGlobals();

   do {
      next_chromosome = engine.SetupMap(next_chromosome);
      many_chromosomes = many_chromosomes || (next_chromosome != 0);

      if (many_chromosomes)
         printf("  Analysing Chromosome %d\n",
                ped.GetMarkerInfo(engine.markers[0])->chromosome);

      // First, calculate the likelihood of each pair using one of the test pedigrees
      for (int i = 0; i < testCases.familyCount; i++)
         {
         Person * test1 = testCases.FindPerson(testCases.families[i]->famid, "TEST1");
         Person * test2 = testCases.FindPerson(testCases.families[i]->famid, "TEST2");

         if (test1 == NULL || test2 == NULL)
            error("Every family in test case pedigree must have individuals named test1 and test2");

         for (int j = 0, pair = 0; j < ped.count; j++)
            {
            if (ped[j].ngeno == 0) continue;

            int last = !pl.withinFamilyPairs ? ped.count :
                       ped.FindFamily(ped[j].famid)->last + 1;

            for (int k = j + 1; k < last; k++)
               {
               if (ped[k].ngeno == 0) continue;
               if (pair < jobStart)  {pair++; continue; }
               if (pair >= jobEnd)   break;

               test1->sex = ped[j].sex;
               test2->sex = ped[j].sex;

               for (int m = 0; m < ped.markerCount; m++)
                  {
                  test1->markers[m] = ped[j].markers[m];
                  test2->markers[m] = ped[k].markers[m];
                  }
               test1->ngeno = ped[j].ngeno;
               test2->ngeno = ped[k].ngeno;

               if (engine.SelectFamily(testCases.families[i]))
                  likelihoods[i][pair++] += engine.CalculateLikelihood();
               }
            }
         }
   } while (next_chromosome);
   printf("\n");

   engine.CleanupGlobals();

   Family * unrelatedFamily = testCases.FindFamily("Unrelated");

   if (unrelatedFamily == NULL)
      error("Every test case pedigree set should include a pair of unrelated individuals");

   int unrelated = unrelatedFamily->serial;

   for (int j = 0, pair = 0; j < ped.count; j++)
      {
      if (ped[j].ngeno == 0) continue;

      int last = !pl.withinFamilyPairs ? ped.count :
                 ped.FindFamily(ped[j].famid)->last + 1;

      for (int k = j + 1; k < last; k++)
         {
         if (ped[k].ngeno == 0) continue;
         if (pair < jobStart)  {pair++; continue; }
         if (pair >= jobEnd)   break;

         if (ped[j].famid != ped[k].famid)
            {
            likelihoods[testCases.familyCount][pair] = likelihoods[unrelated][pair];
            pair++;
            continue;
            }

         Family * family = ped.FindFamily(ped[j].famid);

         Pedigree dummyPed;

         for (int i = family->first; i <= family->last; i++)
            dummyPed.AddPerson(ped[i].famid, ped[i].pid, ped[i].fatid, ped[i].motid, ped[i].sex, true);
         dummyPed.Sort();

         Person * test1 = dummyPed.FindPerson(ped[j].famid, ped[j].pid);
         Person * test2 = dummyPed.FindPerson(ped[k].famid, ped[k].pid);

         for (int m = 0; m < ped.markerCount; m++)
            {
            test1->markers[m] = ped[j].markers[m];
            test2->markers[m] = ped[k].markers[m];
            }
         test1->ngeno = ped[j].ngeno;
         test2->ngeno = ped[k].ngeno;
         test1->zygosity = ped[j].zygosity;
         test2->zygosity = ped[k].zygosity;

         dummyPed.Trim(true);

         if (dummyPed.count <= 1 ||
             2 * dummyPed.families[0]->nonFounders - dummyPed.families[0]->founders > 20)
            {
            likelihoods[testCases.familyCount][pair] = likelihoods[unrelated][pair];
            pair++;
            continue;
            }

         MerlinCore engine(dummyPed);

         engine.SetupGlobals();

         next_chromosome = -2;
         do {
            next_chromosome = engine.SetupMap(next_chromosome);

            engine.SelectFamily(dummyPed.families[0]);

            likelihoods[testCases.familyCount][pair] += engine.CalculateLikelihood();
         } while (next_chromosome);

         engine.CleanupGlobals();

         pair++;
         }
      }

   int mismatches = 0;
   printf("Comparing likelihoods to originally reported likelihood\n");
   for (int j = 0, pair = 0; j < ped.count; j++)
      {
      if (ped[j].ngeno == 0) continue;

      int last = !pl.withinFamilyPairs ? ped.count :
                 ped.FindFamily(ped[j].famid)->last + 1;

      for (int k = j + 1; k < last; k++)
         {
         if (ped[k].ngeno == 0) continue;
         if (pair < jobStart)  {pair++; continue; }
         if (pair >= jobEnd)   break;

         int best_configuration = testCases.familyCount;
         double best_likelihood = likelihoods[testCases.familyCount][pair];

         for (int i = 0; i < testCases.familyCount; i++)
            if (likelihoods[i][pair] > best_likelihood)
               {
               best_likelihood = likelihoods[i][pair];
               best_configuration = i;
               }

         if (best_configuration != testCases.familyCount)
            {
            printf("  %s->%s and %s->%s are %s [llk = %.2f]\n",
                (const char *) ped[j].famid, (const char *) ped[j].pid,
                (const char *) ped[k].famid, (const char *) ped[k].pid,
                (const char *) testCases.families[best_configuration]->famid,
                best_likelihood - likelihoods[testCases.familyCount][pair]);

            // for (int i = 0; i < testCases.familyCount; i++)
            //   printf("   [%s]\t%.2f\n",
            //            (const char *) testCases.families[i]->famid,
            //            likelihoods[i][pair]);
            // printf("   [ORIGINAL]\t%.2f\n", likelihoods[testCases.familyCount][pair]);

            mismatches++;
            }

         pair++;
         }
      }

   printf("\n");
   printf("Found %d potentially misspecified relationships\n", mismatches);
   }

void ShowBanner()
   {
#ifndef VERSION
   printf("RELATIVE FINDER - (c) 2010 Goncalo Abecasis");
#else
   printf("RELATIVE FINDER " VERSION " - (c) 2010 Goncalo Abecasis");
#endif

   printf("\n\n");
   }



 
